# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas

from mozdetect.detectors.base import BaseDetector


class CDFSquaredDetector(BaseDetector):
    """Uses the CDF of two groups of histograms to detect changes."""

    def detect_changes(self, groups=None, **kwargs):
        """Obtain the difference between two CDFs using a squared diff method.

        Two data points are expected in the groups, with a CDF defined in both of them.
        """
        current_date_hist = groups[0]
        next_date_hist = groups[1]

        merged_hist = pandas.merge(
            current_date_hist, next_date_hist, on="bin", suffixes=("_current", "_next")
        )
        merged_hist["diff"] = merged_hist["cdf_current"] - merged_hist["cdf_next"]
        merged_hist["sq_diff"] = (merged_hist["cdf_next"] - merged_hist["cdf_current"]) ** 2

        return {
            "diff": [merged_hist["diff"].sum()],
            "sq_diff": [
                merged_hist["sq_diff"].sum()
                * ((merged_hist["diff"].sum()) / abs(merged_hist["diff"].sum()))
            ],
            "maxmin": (abs(merged_hist["diff"].min()) + abs(merged_hist["diff"].max())),
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas


class BaseTimeSeriesDetector:
    """Base timeseries detector that detectors must inherit from."""

    def __init__(self, timeseries, **kwargs):
        """Initialize the BaseTimeSeriesDetector.

        :param TimeSeries timeseries: A TimeSeries object that represents
            the timeseries to analyze.
        """
        self.timeseries = timeseries
        if hasattr(self.timeseries, "set_data_type"):
            self.timeseries.set_data_type("numerical")

    def get_sum_of_previous_n(self, n, inclusive=False):
        """Returns the sum of the past n data points.

        :param int n: The number of data points to sum.
        :param bool inclusive: If true, include the current point in the sum.
        :return DataFrame: A single data point as the sum of the previous n data points.
        """
        previous_n = self.timeseries.get_previous_n(n, inclusive=inclusive)
        if previous_n.empty:
            return previous_n
        return pandas.DataFrame(previous_n.sum()).T

    def get_sum_of_next_n(self, n, inclusive=True):
        """Returns the sum of the past n data points.

        :param int n: The number of data points to sum.
        :param bool inclusive: If true, include the current point in the sum.
        :return DataFrame: A single data point as the sum of the next n data points.
        """
        next_n = self.timeseries.get_next_n(n, inclusive=inclusive)
        if next_n.empty:
            return next_n
        return pandas.DataFrame(next_n.sum()).T

    def get_avg_of_previous_n(self, n, inclusive=False):
        """Returns the average of the past n data points.

        :param int n: The number of data points to average.
        :param bool inclusive: If true, include the current point in the average.
        :return DataFrame: A single data point as the average of the previous n data points.
        """
        previous_n = self.timeseries.get_previous_n(n, inclusive=inclusive)
        if previous_n.empty:
            return previous_n
        return pandas.DataFrame(previous_n.mean()).T

    def get_avg_of_next_n(self, n, inclusive=True):
        """Returns the average of the next n data points.

        :param int n: The number of data points to sum.
        :param bool inclusive: If true, include the current point in the average.
        :return DataFrame: A single data point as the average of the next n data points.
        """
        next_n = self.timeseries.get_next_n(n, inclusive=inclusive)
        if next_n.empty:
            return next_n
        return pandas.DataFrame(next_n.mean()).T

    def get_avg_of_surrounding_n(self, n):
        pn = int(n / 2)
        if n % 2 == 1:
            nn = pn + 1
        else:
            nn = pn

        currind = self.timeseries._currind
        data = self.timeseries.data.iloc[currind - pn : currind + nn + 1]
        return pandas.DataFrame(data.mean()).T

    def get_avg_of_next_surrounding_n(self, n):
        pn = int(n / 2)
        if n % 2 == 1:
            nn = pn + 1
        else:
            nn = pn

        currind = self.timeseries._currind
        data = self.timeseries.data.iloc[currind - pn + 1 : currind + nn + 2]
        return pandas.DataFrame(data.mean()).T

    def detect_changes(self, **kwargs):
        """Detect changes in a timeseries.

        :return: A list of Detection objects representing the regressions found.
        """
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest

from mozdetect.data import InvalidNumberError, TimeSeries, UnknownDataTypeError


def test_data():
    ts = TimeSeries([(1, 2, 3), (4, 5, 6)])

    # Check that the first row is returned
    assert (ts.get_current() == (1, 2, 3)).all(axis=1).any()

    # Check second time to make sure it doesn't change
    assert (ts.get_current() == (1, 2, 3)).all(axis=1).any()


def test_data_empty():
    ts = TimeSeries([])
    assert len(ts.data) == 0

    with pytest.raises(IndexError):
        ts.get_current()


def test_data_iteration():
    data = [(1, 2, 3), (4, 5, 6)]
    ts = TimeSeries(data)

    for row in ts:
        assert (row == data[ts._currind]).all().any()


def test_data_complex():
    ts = TimeSeries([(1, 2, "h"), (4, 5, "e", 5)])
    assert (ts.get_current() == (1, 2, "h", None)).all().any()
    assert ts.get_current().isna().loc[0, 3]


def test_data_get_n():
    ts = TimeSeries([(1, 2, 3), (4, 5, 6), (3, 5, 5)])

    subset = ts.get_next_n(2)
    assert all((subset == row).all(axis=1).any() for row in [(4, 5, 6), (3, 5, 5)])
    assert not (subset == (1, 2, 3)).all(axis=1).any()

    # Returns empty as there is nothing before the first element
    subset = ts.get_previous_n(1)
    assert subset.empty

    ts._currind += 2
    subset = ts.get_previous_n(2)
    assert all((subset == row).all(axis=1).any() for row in [(1, 2, 3), (4, 5, 6)])
    assert not (subset == (3, 5, 5)).all(axis=1).any()

    # Returns empty as there is nothing after the last element
    subset = ts.get_next_n(1)
    assert subset.empty


def test_data_get_bad_n():
    ts = TimeSeries([])
    with pytest.raises(InvalidNumberError):
        ts.get_next_n(-1)

    with pytest.raises(InvalidNumberError):
        ts.get_previous_n(-1)


def test_data_get_data_type():
    ts = TimeSeries([(1, 2, "a"), (1, 2, "b"), (1, 2, "c")])

    subset = ts.get_numerical_columns()
    assert all((subset == row).all(axis=1).any() for row in [(1, 2), (1, 2), (1, 2)])

    subset = ts.get_nonnumerical_columns()
    assert all((subset == row).all(axis=1).any() for row in [("a",), ("b",), ("c",)])


def test_data_set_data_type():
    ts = TimeSeries([(1, 2, 3.0, "a"), (1, 2, 3.0, "a"), (1, 2, 3.0, "a")])

    ts.set_data_type("numerical")
    for row in ts:
        assert (row == (1, 2, 3.0)).all().any()

    ts.set_data_type("non-numerical")
    for row in ts:
        assert (row == ("a")).all().any()

    ts.set_data_type([int])
    for row in ts:
        assert (row == (1, 2)).all().any()

    ts.set_data_type("all")
    assert (ts.data == (1, 2, 3.0, "a")).all(axis=1).all()


def test_data_set_data_type_failues():
    ts = TimeSeries([(1, 2, 3.0, "a"), (1, 2, 3.0, "a"), (1, 2, 3.0, "a")])

    with pytest.raises(UnknownDataTypeError):
        ts.set_data_type("bad-type")

    with pytest.raises(UnknownDataTypeError):
        ts.set_data_type(["not a type"])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import json
import numpy as np
import pandas
import traceback

from collections import Counter
from datetime import datetime, timedelta


class InvalidNumberError(Exception):
    """Raised when an invalid number of points is requested."""

    pass


class UnknownDataTypeError(Exception):
    """Raised when an unknown data type is requested."""

    pass


class TimeSeries:
    """Represents a time series composed of Datum objects.

    Primarily a wrapper around `pandas.DataFrame`. It provides some additional
    helper functions to make it simpler to iterate over the data, and provide
    a common interface for all detectors to use.

    The pandas.DataFrame was chosen for this so that strings could be included
    in the dataset alongside numerical input. It also easily allows multidimensional
    data to be used.
    """

    def __init__(self, data, data_type="all", **kwargs):
        """Initializes the time series.

        :param list data: A list of tuples representing the time series.
        :param str/list data_type: The data type that should be iterated over. See
            `set_data_type` for more information.
        :param kwargs: A set of options that can be used to finetune the
            options passed into the pandas DataFrame creation.
        """
        self._original_data = self._prepare_data(data, **kwargs)
        self._iteration_data = self._original_data
        self._data_type = "all"

        self.numerical_data = self.get_numerical_columns()
        self.nonnumerical_data = self.get_nonnumerical_columns()

        # Only set data type after getting the numberical, and
        # non-numerical columns from the original data
        self.set_data_type(data_type)

        self._currind = 0
        self._curr = None

    def __iter__(self):
        """Helper method to iterate over the timeseries.

        :return DataFrame: A row in the DataFrame.
        """
        for index, row in self.data.iterrows():
            self._currind = index
            self._curr = pandas.DataFrame(row).T

            yield self._curr

    @property
    def data(self):
        return self._iteration_data

    def _prepare_data(self, data, **kwargs):
        """Formats the data into a pandas DataFrame for detectors.

        Note that since the DataFrame doesn't have restrictions on types,
        multiple types can be combined into a single tuple, e.g. (1, 2, "h")
        is valid. As is the following when data is missing: [(1, 2), (3, 4, 5)]

        :param list data: A list of tuples representing the time series.
        :param kwargs: A set of options that can be used to finetune the
            options passed into the pandas DataFrame creation.

        :return DataFrame: The data converted to a pandas DataFrame.
        """
        return pandas.DataFrame(data=data, **kwargs)

    def set_data_type(self, data_type):
        """Used to set the data type to iterate over.

        By default, all the data will be iterated over. If "numerical" is passed
        here, then only the numerical data will be returned in all methods. If
        "non-numerical" is passed, then only the non-numerical data will be returned.
        This can be reset to all data by passing "all". A custom type may also
        be passed for special returns.

        :param str/list data_type: Either "numerical", "non-numerical", or "all" to
            denote the type of data. Alternatively, pass a list of custom types to get
            alternative data.
        """
        self._data_type = data_type
        if isinstance(self._data_type, str):
            if self._data_type == "numerical":
                self._iteration_data = self.numerical_data
            elif self._data_type == "non-numerical":
                self._iteration_data = self.nonnumerical_data
            elif self._data_type == "all":
                self._iteration_data = self._original_data
            else:
                raise UnknownDataTypeError(
                    f"Unknown data type requested for iteration: {self._data_type}"
                )
        elif isinstance(self._data_type, list):
            try:
                self._iteration_data = self._original_data.select_dtypes(include=self._data_type)
            except Exception:
                raise UnknownDataTypeError(
                    f"Failed to get custom data types for {str(self._data_type)}:"
                    f" {traceback.format_exc()}"
                )
        else:
            raise UnknownDataTypeError("Expecting list or str as type for data type.")

    def get_current(self):
        """Returns the current datum being analyzed.

        :return DataFrame: The row at the current position in the DataFrame.
        """
        if self._curr is None:
            self._curr = self.data.iloc[[self._currind]]
        return self._curr

    def get_next_n(self, n, inclusive=False):
        """Returns the next `n` data points in the time series.

        This methods is exclusive, and doesn't include the current data point
        that is being analyzed.

        :param int n: The number of data points to get.

        :return DataFrame: The number of requested rows if they exist. If the
            current position is at the end of the timeseries, then nothing will
            be returned.
        """
        if n <= 0:
            raise InvalidNumberError("Number of data points must be greater than 0.")

        start_ind = self._currind + 1
        end_ind = self._currind + n + 1
        if inclusive:
            start_ind = start_ind - 1
            end_ind = end_ind - 1

        return self.data.iloc[start_ind:end_ind]

    def get_previous_n(self, n, inclusive=False):
        """Returns the previous `n` data points in the time series.

        This methods is exclusive, and doesn't include the current data point
        that is being analyzed.

        :param int n: The number of data points to get.

        :return DataFrame: The number of requested rows if they exist. If the
            current position is at the beginning of the timeseries, then nothing will
            be returned.
        """
        if n <= 0:
            raise InvalidNumberError("Number of data points must be greater than 0.")

        start_ind = max(self._currind - n, 0)
        end_ind = self._currind
        if inclusive:
            if self._currind == 0:
                start_ind = 0
            else:
                start_ind = start_ind + 1
            end_ind = end_ind + 1

        return self.data.iloc[start_ind:end_ind]

    def get_numerical_columns(self):
        """Returns the data but with only numerical columns.

        :return DataFrame: Returns the data with non-numerical columns removed.
        """
        return self.data.select_dtypes(include=[np.number])

    def get_nonnumerical_columns(self):
        """Returns the data but with only non-numerical columns.

        :return DataFrame: Returns the data with numerical columns removed.
        """
        return self.data.select_dtypes(exclude=[np.number])


class TelemetryTimeSeries:
    """Provides additional telemetry-specific methods, and exposes the
    raw data in the `raw_data` attribute to provide more customization.
    """

    def __init__(self, data, *args, **kwargs):
        self.raw_data = data
        self.cumulative_by_day_histograms = pandas.DataFrame()
        self.cumulative_multiday_histograms = pandas.DataFrame()

    def _build_id_to_date(self, build_id):
        """Returns a build ID converted into a datetime.date.

        :param str build_id: The build ID to convert.

        :return datetime.date: The date that was parsed from the build ID.
        """
        if len(build_id) == 10:
            return datetime.strptime(build_id, "%Y%m%d%H").date()
        else:
            return datetime.strptime(build_id, "%Y%m%d%H%M%S").date()

    def get_percentiles_over_time(self, histograms=None):
        """Returns the percentiles over time from the telemetry data."""

        def _extract_percentiles(group):
            filtered = group[group["cdf"] > 0.5]
            retval = filtered.iloc[0][["date"]]
            retval["p50"] = filtered.iloc[0]["bin"]
            retval["p75"] = group[group["cdf"] > 0.75].iloc[0]["bin"]
            retval["p95"] = group[group["cdf"] > 0.95].iloc[0]["bin"]
            return retval

        if histograms is None:
            if not self.cumulative_by_day_histograms.empty:
                histograms = self.cumulative_by_day_histograms
            else:
                histograms = self.raw_data

        return (
            histograms.groupby("date", as_index=False)[["date", "bin", "cdf"]]
            .apply(_extract_percentiles)
            .dropna()
            .reset_index(drop=True)
        )

    def get_cumulative_by_day(self, start_date=None, end_date=None):
        """Returns the data downsampled into a per-day granularity.

        :param datetime.date start_date: The start date to downsample from.
        :param datetime.date end_date: The end date to downsample to.

        :return pandas.DataFrame: A pandas.DateFrame of the cumulative histograms at a per-day
            granularity. Can also be obtained from `self.cumulative_histograms`.
        """
        last_start_row = 0
        current_date = start_date or self._build_id_to_date(self.raw_data.iloc[0].build_id)
        end_date = end_date or self._build_id_to_date(self.raw_data.iloc[-1].build_id)

        while current_date <= end_date:
            # Find the start date location
            # TODO: get rid of while True loops
            while True:
                current_row = self.raw_data.iloc[last_start_row]
                current_row_date = self._build_id_to_date(current_row.build_id)
                if current_row_date < current_date:
                    last_start_row += 1
                else:
                    break

            # Gather the data from builds in the current day
            summed_histograms_raw = Counter()
            current_row_iloc = last_start_row
            while current_row_iloc < len(self.raw_data):
                current_row = self.raw_data.iloc[current_row_iloc]
                current_row_date = self._build_id_to_date(current_row.build_id)
                if current_row_date > current_date:
                    break
                if not current_row.non_norm_histogram:
                    current_row_iloc += 1
                    continue

                summed_histograms_raw.update(json.loads(current_row.non_norm_histogram))
                current_row_iloc += 1

            summed_histogram = pandas.DataFrame(
                list(summed_histograms_raw.items()), columns=["bin", "count"]
            )
            if summed_histogram.empty:
                current_date += timedelta(days=1)
                continue

            # Produce the cumulative histogram for this day
            summed_histogram["bin"] = summed_histogram["bin"].astype(int)
            summed_histogram["date"] = current_date

            summed_histogram = summed_histogram.sort_values(by="bin").reset_index(drop=True)
            summed_histogram["cumulative"] = summed_histogram["count"].cumsum()
            summed_histogram["cdf"] = (
                summed_histogram["cumulative"] / summed_histogram["count"].sum()
            )

            self.cumulative_by_day_histograms = pandas.concat(
                [summed_histogram.fillna(0), self.cumulative_by_day_histograms], ignore_index=True
            )
            current_date += timedelta(days=1)

        return self.cumulative_by_day_histograms

    def get_multiday_average(self, days=7):
        """Produces a multiday average of the cumulative per-day histograms using
        a rolling window.

        :param int days: The number of days to average together.

        :return pandas.DataFrame: The data with each point being a multiday average instead
            of only a single day.
        """
        if self.cumulative_by_day_histograms is None:
            raise Exception(
                "get_multiday_average expects get_cumulative_by_day to be called first."
            )

        start_date = self.cumulative_by_day_histograms["date"].min()
        end_date = self.cumulative_by_day_histograms["date"].max()

        # Initialize the current sum with the start date
        current_date = start_date
        current_sum = self.cumulative_by_day_histograms[
            self.cumulative_by_day_histograms["date"] == current_date
        ][["bin", "count"]].reset_index(drop=True)

        # Gather the first set of days-1 into a sum
        # TODO use walrus here
        current_date += timedelta(days=1)
        while current_date < start_date + timedelta(days=days - 1):
            current_sum["count"] += self.cumulative_by_day_histograms[
                self.cumulative_by_day_histograms["date"] == current_date
            ][["bin", "count"]].reset_index(drop=True)["count"]
            current_date += timedelta(days=1)

        # Add one day at a time to the current_sum, and remove the last day from it
        # after adding the current_sum to the multiday_window data
        # TODO use walrus here
        current_date = start_date
        date_to_add = current_date + timedelta(days=days - 1)
        while date_to_add <= end_date:
            # Add new date data if it has data
            hist_to_add = self.cumulative_by_day_histograms[
                self.cumulative_by_day_histograms["date"] == date_to_add
            ][["bin", "count"]].reset_index(drop=True)["count"]
            if not hist_to_add.empty:
                current_sum["count"] += hist_to_add

            # Calculate new CDF, and add to multiday_window
            current_sum["date"] = current_date
            current_sum["cumulative"] = current_sum["count"].cumsum()
            current_sum["cdf"] = current_sum["cumulative"] / current_sum["count"].sum()
            self.cumulative_multiday_histograms = pandas.concat(
                [current_sum.copy(), self.cumulative_multiday_histograms], ignore_index=True
            )

            # Remove current date from current_sum if it had data
            hist_to_remove = self.cumulative_by_day_histograms[
                self.cumulative_by_day_histograms["date"] == current_date
            ][["bin", "count"]].reset_index(drop=True)["count"]
            if not hist_to_remove.empty:
                current_sum["count"] -= hist_to_remove

            current_date += timedelta(days=1)
            date_to_add += timedelta(days=1)

        return self.cumulative_multiday_histograms



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


class BaseDetector:
    """Base class for all group detectors."""

    def __init__(self, groups=None, **kwargs):
        """Initialize the detector.

        :param list groups: A list of DataFrame objects to compare between.
            Generally expected for there to be TWO groups to compare, but it's
            possible to have multiple to do a cross-comparison (assuming the
            detector supports this).
        """
        self.groups = groups

    def _coalesce_groups(self, groups):
        """Used to determine the groups to compare.

        The groups passed as an argument have a higher priority than the groups
        used to initialize the detector.

        :param list groups: A list of groups to compare or None.
        :return list: The groups that should be compared.
        """
        if not groups:
            if not self.groups:
                raise ValueError("Groups to compare have not been specified.")
            return self.groups
        return groups

    def detect_changes(self, groups=None, **kwargs):
        """Detect changes between two groups of data points.

        :param list groups: A list of the groups to compare.
        """
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozdetect.data import TelemetryTimeSeries, TimeSeries
from mozdetect.detectors import get_detectors
from mozdetect.timeseries_detectors import get_timeseries_detectors

__all__ = [
    "get_detectors",
    "get_timeseries_detectors",
    "TelemetryTimeSeries",
    "TimeSeries",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from datetime import datetime
from google.cloud import bigquery


class BigQueryClient:
    client = None

    def __init__(self, project="moz-fx-data-bq-performance"):
        if BigQueryClient.client is None:
            BigQueryClient.initialize_bq_client()

    @classmethod
    def initialize_bq_client(cls, project="moz-fx-data-bq-performance"):
        BigQueryClient.client = bigquery.Client()


def get_years_to_query():
    current_year = datetime.now().year
    next_year = current_year + 1
    previous_year = current_year - 1
    return previous_year, current_year, next_year


def _get_android_metric_table(probe):
    previous_year, current_year, next_year = get_years_to_query()

    job = BigQueryClient.client.query(
        f"""
        SELECT *
        FROM
            moz-fx-data-shared-prod.glam_etl.glam_fenix_nightly_aggregates
        WHERE
            metric = '{probe}'
            AND ping_type = 'metrics'
            AND os = 'Android'
            AND build_id != '*'
            AND (
                build_id like '{previous_year}%'
                OR build_id like '{current_year}%'
                OR build_id like '{next_year}%'
            )
    """
    )
    return job.to_dataframe()


def _get_non_fog_desktop_metric_table(probe, process, os):
    previous_year, current_year, next_year = get_years_to_query()

    job = BigQueryClient.client.query(
        f"""
        SELECT *
        FROM
            moz-fx-data-shared-prod.glam_etl.glam_desktop_nightly_aggregates
        WHERE
            metric ='{probe}'
            AND process = "{process}"
            AND os = '{os}'
            AND build_id != "*"
            AND (
                build_id like '{previous_year}%'
                OR build_id like '{current_year}%'
                OR build_id like '{next_year}%'
            )
        ORDER BY build_id
    """
    )

    return job.to_dataframe()


def _get_fog_desktop_metric_table(probe, os):
    previous_year, current_year, next_year = get_years_to_query()

    job = BigQueryClient.client.query(
        f"""
        SELECT *
        FROM
            moz-fx-data-shared-prod.glam_etl.glam_fog_nightly_aggregates
        WHERE
            metric ='{probe}'
            AND ping_type = "*"
            AND os = '{os}'
            AND build_id != "*"
            AND (
                build_id like '{previous_year}%'
                OR build_id like '{current_year}%'
                OR build_id like '{next_year}%'
            )
        ORDER BY build_id
    """
    )
    return job.to_dataframe()


def get_metric_table(probe, os, process=None, android=False, use_fog=False):
    BigQueryClient()

    print("Running query...")
    if android:
        return _get_android_metric_table(probe, os, process)
    elif not use_fog:
        if process is None:
            raise ValueError("Missing process argument for non-fog telemetry probes.")
        return _get_non_fog_desktop_metric_table(probe, process, os)
    else:
        return _get_fog_desktop_metric_table(probe, os)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np


from mozdetect.detectors.base import BaseDetector


class CDFDetector(BaseDetector):
    """Uses the CDF of two groups of histograms to detect changes."""

    def _calculate_cdf(self, group):
        cdf = []
        currsum = 0
        for col, values in group.items():
            currsum += values[0]
            cdf.append(currsum)
        cdf = list((np.asarray(cdf) / cdf[-1]) * 100)
        return cdf

    def _calculate_cdf_difference(self, cdf_a, cdf_b):
        diff = 0
        for val_a, val_b in zip(cdf_a, cdf_b):
            diff += abs(val_a - val_b)
        return diff

    def _calculate_cdf_area_difference(self, cdf_a, cdf_b):
        z = np.asarray(cdf_a) - np.asarray(cdf_b)
        x = np.arange(len(cdf_a))

        # Determine intersection points of the CDFs
        dx = x[1:] - x[:-1]
        cross_test = np.sign(z[:-1] * z[1:])
        dx_intersect = -dx / (z[1:] - z[:-1]) * z[:-1]

        # Find areas of positive, and negative areas
        areas_pos = abs(z[:-1] + z[1:]) * 0.5 * dx
        areas_neg = 0.5 * dx_intersect * abs(z[:-1]) + 0.5 * (dx - dx_intersect) * abs(z[1:])

        # Find the total area between the curves
        areas = np.where(cross_test < 0, areas_neg, areas_pos)
        total_area = np.sum(areas)

        return total_area

    def detect_changes(self, groups=None, **kwargs):
        groups = self._coalesce_groups(groups)
        group_a = groups[0]
        group_b = groups[1]

        cdf_a = self._calculate_cdf(group_a)
        cdf_b = self._calculate_cdf(group_b)

        diff = self._calculate_cdf_difference(cdf_a, cdf_b)

        total_diff = len(cdf_a) * 100
        diff_pct = (diff / total_diff) * 100

        return diff, diff_pct, total_diff



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


class Detection:
    """Used to represent a change detection in a timeseries."""

    def __init__(self, previous_value, new_value, confidence, location, direction):
        """Create a new detected change.

        :param float previous_value: The previous value before the detected change.
        :param float new_value: The new value at the detected change location.
        :param float confidence: The confidence of the detected change.
        :param str location: The location of the detected change. Can be a revision,
            build_id, or something else that represents the place at which a regression
            was detected.
        """
        self.previous_value = previous_value
        self.new_value = new_value
        self.confidence = confidence
        self.location = location
        self.direction = direction

    def __repr__(self):
        return (
            f"Detection<"
            f"previous_value={self.previous_value}, new_value={self.new_value}, "
            f"confidence={self.confidence}, location={self.location}, "
            f"direction={self.direction}>"
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozdetect.detectors import CDFDetector
from mozdetect.timeseries_detectors.base import BaseTimeSeriesDetector
from mozdetect.timeseries_detectors.detection import Detection
from mozdetect.utils import is_dev_mode


class CDFTimeSeriesDetector(BaseTimeSeriesDetector):
    """Analyzes a time series of histograms using CDFs to detect changes."""

    def detect_changes(self, group_size=7, alert_threshold=0.1, **kwargs):
        detections = []
        detector = CDFDetector()

        prev_detection = None
        cdf_diffs = []
        build_ids = self.timeseries.nonnumerical_data
        max_detection = None
        for i, row in enumerate(self.timeseries):
            if i < group_size or i > (len(build_ids["build_id"]) - group_size):
                # Ensure that we're always comparing groups
                # that have the same size
                continue
            build_id = build_ids["build_id"][i]
            avg_hist_previous = self.get_sum_of_previous_n(group_size, inclusive=False)
            avg_hist_current = self.get_sum_of_next_n(group_size, inclusive=True)

            (cdf_diff, cdf_diff_pct, cdf_diff_total) = detector.detect_changes(
                [avg_hist_previous, avg_hist_current]
            )
            cdf_diffs.append(cdf_diff_pct)

            if cdf_diff_pct > alert_threshold and prev_detection:
                if not max_detection:
                    max_detection = Detection(cdf_diff, prev_detection[0], cdf_diff_pct, build_id)
                elif cdf_diff_pct > max_detection.confidence:
                    max_detection = Detection(cdf_diff, prev_detection[0], cdf_diff_pct, build_id)
                detections.append(Detection(cdf_diff, prev_detection[0], cdf_diff_pct, build_id))
            prev_detection = (cdf_diff, cdf_diff_pct, cdf_diff_total)

        if is_dev_mode():
            from matplotlib import pyplot as plt

            plt.figure()
            plt.plot(cdf_diffs)
            plt.xticks(
                ticks=list(range(len(cdf_diffs))),
                labels=list(build_ids["build_id"])[group_size : -group_size + 1],
            )

            plt.show()

        if max_detection:
            return detections, [max_detection]
        return detections, []



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from importlib import util as importutil
from scipy.signal import butter, filtfilt


def butter_lowpass(cutoff, fs, order=5):
    nyquist = 0.5 * fs
    normal_cutoff = cutoff / nyquist
    b, a = butter(order, normal_cutoff, btype="low", analog=False)
    return b, a


def lowpass_filter(data, cutoff, fs, order=5):
    b, a = butter_lowpass(cutoff, fs, order=order)
    y = filtfilt(b, a, data)
    return y


def is_dev_mode():
    return importutil.find_spec("matplotlib") is not None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozdetect.timeseries_detectors.base import BaseTimeSeriesDetector
from mozdetect.timeseries_detectors.cdf import CDFTimeSeriesDetector
from mozdetect.timeseries_detectors.cdf_squared import CDFSquaredTimeSeriesDetector

TIMESERIES_DETECTORS = {
    "base": BaseTimeSeriesDetector,
    "cdf": CDFTimeSeriesDetector,
    "cdf_squared": CDFSquaredTimeSeriesDetector,
}


def get_timeseries_detectors():
    return TIMESERIES_DETECTORS



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas

from datetime import timedelta
from prettytable import PrettyTable

from mozdetect.detectors import CDFSquaredDetector
from mozdetect.timeseries_detectors.base import BaseTimeSeriesDetector
from mozdetect.timeseries_detectors.detection import Detection
from mozdetect.utils import lowpass_filter


class CDFSquaredTimeSeriesDetector(BaseTimeSeriesDetector):
    """Analyzes a time series of histograms using CDFs to detect changes."""

    def __init__(self, timeseries, start_date=None, end_date=None, **kwargs):
        super().__init__(timeseries, **kwargs)
        self.start_date = start_date
        self.end_date = end_date

    def _coalesce_dates(self, start_date, end_date):
        return (
            self.start_date or start_date,
            self.end_date or end_date,
        )

    def _calculate_differences(self):
        """Calculates all the CDF differences across the full time series."""
        detector = CDFSquaredDetector()

        start_date, end_date = self._coalesce_dates(
            self.timeseries.cumulative_multiday_histograms["date"].min(),
            self.timeseries.cumulative_multiday_histograms["date"].max(),
        )

        current_date = start_date
        differences = pandas.DataFrame()
        while current_date <= end_date - timedelta(days=15):
            current_date_hist = self.timeseries.cumulative_multiday_histograms[
                self.timeseries.cumulative_multiday_histograms["date"] == current_date
            ][["bin", "cdf"]]
            next_date_hist = self.timeseries.cumulative_multiday_histograms[
                self.timeseries.cumulative_multiday_histograms["date"]
                == current_date + timedelta(days=7)
            ][["bin", "cdf"]]

            difference = detector.detect_changes(groups=[current_date_hist, next_date_hist])
            difference["date"] = [current_date + timedelta(days=6)]
            differences = pandas.concat(
                [differences, pandas.DataFrame(difference)], ignore_index=True
            )

            current_date += timedelta(days=1)

        print(differences)
        differences["filtered_sq_diff"] = lowpass_filter(differences["sq_diff"], 20, 100)

        return differences

    def _find_alerts(self, differences, alert_threshold=0.85):
        start_date, end_date = self._coalesce_dates(
            differences["date"].min(), differences["date"].max()
        )
        current_date = start_date + timedelta(days=7)

        threshold = abs(differences["filtered_sq_diff"]).quantile(alert_threshold)

        tracking_diff = False
        found_peak = False

        last_value = None
        detections = pandas.DataFrame()
        while current_date <= end_date - timedelta(days=15):
            diff = differences[differences["date"] == current_date]["filtered_sq_diff"].iloc[0]

            if tracking_diff and abs(diff) > last_value:
                # heading for new peak
                found_peak = False

            if not found_peak and tracking_diff and abs(diff) < last_value:
                if diff > 0:
                    direction = "up"
                else:
                    direction = "down"
                new_row = pandas.DataFrame({"date": [current_date], "direction": [direction]})
                detections = pandas.concat([detections, new_row], ignore_index=True)
                found_peak = True

            if tracking_diff:
                last_value = abs(diff)

            if abs(diff) <= threshold:
                found_peak = False
                tracking_diff = False

            if not tracking_diff and abs(diff) > threshold:
                tracking_diff = True
                last_value = abs(diff)
            current_date += timedelta(days=1)

        return detections

    def _get_interpolated_percentile(self, histogram, percentile):
        """Get an interpolated percentile value from the given histogram.

        :param pandas.DataFrame histogram: The histogram to get an interpolated
            percentile value from.
        :param float percentile:  The percentile to obtain a value for.
        """
        bucket_before = histogram[histogram["cdf"] <= percentile].max()
        bucket = histogram[histogram["cdf"] > percentile].min()
        bucket_after = histogram[histogram["cdf"] > percentile].iloc[1]

        x = np.array([bucket_before["cdf"], bucket["cdf"]])
        y = np.array([bucket["bin"], bucket_after["bin"]])
        return np.interp(percentile, x, y)

    def _describe_detection(self, detection):
        """Produce information about the detection."""
        if self.timeseries.cumulative_multiday_histograms is None:
            raise ValueError(
                "Cannot produce a description of the detection without a multiday average."
            )

        before_histogram = self.timeseries.cumulative_multiday_histograms[
            self.timeseries.cumulative_multiday_histograms["date"]
            == detection["date"] - timedelta(days=7)
        ]
        after_histogram = self.timeseries.cumulative_multiday_histograms[
            self.timeseries.cumulative_multiday_histograms["date"] == detection["date"]
        ]

        merged_hist = pandas.merge(
            before_histogram, after_histogram, on="bin", suffixes=("_current", "_next")
        )
        merged_hist["diff"] = merged_hist["cdf_current"] - merged_hist["cdf_next"]
        total_diff = merged_hist["diff"].sum()

        detection_info = []
        detection_info.append(
            [
                "Total Samples",
                before_histogram["count"].sum(),
                after_histogram["count"].sum(),
            ]
        )
        detection_info.append(
            [
                "Interpolated Median",
                self._get_interpolated_percentile(before_histogram, 0.5),
                self._get_interpolated_percentile(after_histogram, 0.5),
            ]
        )
        detection_info.append(
            [
                "Interpolated p05",
                self._get_interpolated_percentile(before_histogram, 0.05),
                self._get_interpolated_percentile(after_histogram, 0.05),
            ]
        )
        detection_info.append(
            [
                "Interpolated p95",
                self._get_interpolated_percentile(before_histogram, 0.95),
                self._get_interpolated_percentile(after_histogram, 0.95),
            ]
        )
        detection_info.append(["CDF Diff", "", total_diff])
        detection_info.append(
            [
                "CDF Shift",
                "",
                "Left" if total_diff < -0.05 else "Right" if total_diff > 0.05 else "Mixed",
            ]
        )

        table = PrettyTable()
        table.field_names = ["Metric", "Before", "After"]
        table.add_rows(detection_info)
        print("Alert Generated: " + detection["date"].strftime("%Y-%m-%d"))
        print(table)

        return {d[0]: d[1:] for d in detection_info}

    def detect_changes(self, multiday_average_days=7, alert_threshold=0.85, **kwargs):
        """Detects changes in a telemetry probe using the CDF squared detection method.

        :param int multiday_average_days: The number of days to use in the multiday average.

        :return list: A list of Detection objects representing where a detection occurred.
        """
        self.timeseries.get_cumulative_by_day(start_date=self.start_date, end_date=self.end_date)
        self.timeseries.get_multiday_average(days=multiday_average_days)

        differences = self._calculate_differences()
        pre_detections = self._find_alerts(differences, alert_threshold=alert_threshold)

        detections = []
        for i in range(len(pre_detections)):
            detection = pre_detections.iloc[i]
            detection_info = self._describe_detection(detection)

            detections.append(
                Detection(
                    detection_info["Total Samples"][0],
                    detection_info["Total Samples"][1],
                    detection_info["CDF Diff"][1],
                    detection["date"],
                    detection["direction"][1],
                )
            )

        return detections



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozdetect.detectors.base import BaseDetector
from mozdetect.detectors.cdf import CDFDetector
from mozdetect.detectors.cdf_squared import CDFSquaredDetector

DETECTORS = {"base": BaseDetector, "cdf": CDFDetector, "cdf_squared": CDFSquaredDetector}


def get_detectors():
    return DETECTORS
